package com.adobe.media
{
	import com.adobe.media.EquirectangularProjection;
	import com.adobe.utils.AGALMiniAssembler;
	
	import flash.display.Stage;
	import flash.display.Stage3D;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.Context3DTriangleFace;
	import flash.display3D.Context3DVertexBufferFormat;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.Program3D;
	import flash.display3D.VertexBuffer3D;
	import flash.display3D.textures.VideoTexture;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.geom.Matrix3D;
	import flash.geom.Rectangle;
	import flash.geom.Vector3D;
	import flash.net.NetStream;
	
	public class SphericalVideo extends EventDispatcher
	{
		public static const AVAILABLE:String = "Available";
		public static const PROJECTION_NONE:int = 0;
		public static const PROJECTION_EQUIRECTANGLUAR:int = 1;
		public static const PROJECTION_CUBEMAP:int = 2;
		
		private var texture:VideoTexture;
		private var projection:Projection;
		
		private var stage3D:Stage3D;
		private var context3D:Context3D;
		private var program:Program3D; 
		private var vertexBuffer:VertexBuffer3D; 
		private var indexBuffer:IndexBuffer3D;
		private const VERTEX_SHADER:String =
			"m44 op, va0, vc0\n" +  // copy position to output
			"mov v0, va1"; 			// copy uv to varying variable v0
		private const FRAGMENT_SHADER:String = 
			"tex oc, v0, fs0 <2d, clamp, nomip>\n";
		private var vertexAssembly:AGALMiniAssembler = new AGALMiniAssembler();
		private var fragmentAssembly:AGALMiniAssembler = new AGALMiniAssembler();
		
		// World and View transform related variables
		protected var viewTransform:Matrix3D = new Matrix3D();
		protected var projectionTransform:Matrix3D = new Matrix3D();
		
		private var stageWidth:int;
		private var stageHeight:int;
		private var videoArea:Rectangle = new Rectangle();
		
		private var textureReady:Boolean = false;

		////////////////////////////////////////////////////////////////////////
		public function SphericalVideo(stage:Stage, x:int, y:int, width:int, height:int): void
		{
			if(!Context3D.supportsVideoTexture) return;
		
			stageWidth = width;
			stageHeight = height;
			
			texture = null;
			projection = null;
			context3D = null;
			stage3D = stage.stage3Ds[0];
			
			stage3D.x = x;
			stage3D.y = y;
			
			// Creating a Context3D is not instant so we have to wait for an event telling us when it is ready
			stage3D.addEventListener(Event.CONTEXT3D_CREATE, onContext3DCreated);
			stage3D.requestContext3D();
		}
		
		////////////////////////////////////////////////////////////////////////
		// Attach netstream for video
		public function attachNetStream(s:NetStream):void
		{
			if (texture)
				texture.attachNetStream(s);
			
			// Reset internal variables that need to be newly set for new stream
			projection = null;
			viewTransform.identity();
			projectionTransform.identity();
			textureReady = false;
		}
				
		////////////////////////////////////////////////////////////////////////
		// Set Field Of View
		public function setFOV(fovY:Number, zNear:Number, zFar:Number):void
		{
			var fovX:Number = stageWidth / stageHeight * fovY;
			var xScale:Number = 1.0 / Math.tan((fovX / 2) * Math.PI / 180);
			var yScale:Number = 1.0 / Math.tan((fovY / 2) * Math.PI / 180);
			
			projectionTransform.copyRawDataFrom(Vector.<Number>([
				xScale, 0.0, 0.0, 0.0,
				0.0, yScale, 0.0, 0.0,
				0.0, 0.0, zFar/(zFar-zNear), 1.0,
				0.0, 0.0, (zNear*zFar)/(zNear-zFar), 0.0
			]));
	
			if (textureReady)
				renderFrame();
		}
		
		////////////////////////////////////////////////////////////////////////
		// Set Camera View to change point of view
		public function setCameraView(pitch:Number, yaw:Number, roll:Number):void
		{
			viewTransform.identity();
			viewTransform.appendRotation(roll, Vector3D.Z_AXIS);
			viewTransform.appendRotation(yaw, Vector3D.Y_AXIS);
			viewTransform.appendRotation(pitch, Vector3D.X_AXIS);

			if (textureReady)
				renderFrame();
		}
		
		////////////////////////////////////////////////////////////////////////
		// Set Projection Type. This indicate what type of projection should be used for rendering.
		public function setProjectionType(type:int):void
		{
			if (!context3D || !texture) return;
			
			switch (type)
			{
			case PROJECTION_EQUIRECTANGLUAR:
				projection = new EquirectangularProjection();
				break;
			case PROJECTION_CUBEMAP:
				projection = new CubeProjection();
				break;
			default:
				projection = new Projection();
			}

			if (projection)
			{
				// Compile shaders
				vertexAssembly.assemble(Context3DProgramType.VERTEX, VERTEX_SHADER, false);
				fragmentAssembly.assemble(Context3DProgramType.FRAGMENT, FRAGMENT_SHADER, false);       
				
				// Upload and set shader programs
				program = context3D.createProgram();
				program.upload(vertexAssembly.agalcode, fragmentAssembly.agalcode);
				context3D.setProgram(program);
				
				// Update indices and vertices data from Projection
				indexBuffer = context3D.createIndexBuffer(projection.getIndices().length);
				indexBuffer.uploadFromVector(projection.getIndices(), 0, projection.getIndices().length);
				vertexBuffer = context3D.createVertexBuffer(projection.getVertices().length / Projection.DATA_PER_VERTEX, Projection.DATA_PER_VERTEX);
				vertexBuffer.uploadFromVector(projection.getVertices(), 0, projection.getVertices().length / Projection.DATA_PER_VERTEX);
				
				// Identify vertex data inputs for vertex program
				context3D.setVertexBufferAt(0, vertexBuffer, 0, Context3DVertexBufferFormat.FLOAT_3); // va0 is position
				context3D.setVertexBufferAt(1, vertexBuffer, 3, Context3DVertexBufferFormat.FLOAT_2); // va1 is uv
				
				// Set texture
				context3D.setTextureAt(0, texture);
			}
		}
		
		private function onContext3DCreated(e:Event):void
		{
			context3D = stage3D.context3D;
			
			context3D.setCulling(Context3DTriangleFace.BACK);
			context3D.configureBackBuffer(stageWidth, stageHeight, 0);
			
			texture = context3D.createVideoTexture();
			texture.addEventListener(Event.TEXTURE_READY, onTextureReady);

			if(hasEventListener(AVAILABLE))
				dispatchEvent(new Event(AVAILABLE));
		}
		
		private function onTextureReady(event:Event):void
		{
			textureReady = true;
			renderFrame();
		}
		
		private function renderFrame():void 
		{
			var m:Matrix3D = new Matrix3D();

			m.append(viewTransform);
			m.append(projectionTransform);
			
			context3D.clear(0,0,0,0);
			context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 0, m, true); // vc0 is view and projection
			context3D.drawTriangles(indexBuffer);
			context3D.present();

		}
	}
}